import imp
import sys

import logging

import pdfxcb.json1 as json1


lg=logging


try:
    imp.find_module('zbar')
except ImportError:
    msg = json1.json_msg_module_not_accessible('zbar')
    lg.error(msg)
    lg.info(json1.json_last_log_msg())
    sys.exit(msg)
import zbar


# Image is provided by PIL (python 2) or pillow (python 3; debian
# supplies this as python3-pil)
try:
    imp.find_module('PIL')
except ImportError:
    msg = json1.json_msg_module_not_accessible('PIL')
    lg.error(msg)
    lg.info(json1.json_last_log_msg())
    sys.exit(msg)
from PIL import Image



def barcodeScan(imagePNGPath, scan_region):
    """
    imagePNGPath should be a string defining the location of a PNG
    file. Return None if a barcode was not found. If a barcode was
    found, return a string corresponding to the barcode-encoded data.

    Search within the region defined by SCAN_REGION when SCAN_REGION
    is a list. When SCAN_REGION is a list, it specifies two points as
    [x1,y1,x2,y2]. The two points, (x1,y1) and (x2,y2), are pairs
    (x,y) of percentages (each expressed as a value between 0 and 1.0)
    relative to the dimensions of the image; they define the box
    within which the barcode scan occurs.

    If SCAN_REGION is not a list, the full image is analyzed. If
    analysis of the full image is desirable, do not set SCAN_REGION to
    [0,0,1,1] but instead set it to None or some other non-list value.
    """
    # sanity check(s)
    if not isinstance(scan_region,list):
        scan_region = None
    else:
        for value in scan_region:
            if (value < 0 or value > 1):
                msg = json1.json_msg(999,"insane scan region value",False,None)
                lg.error(msg)
                lg.info(json1.json_last_log_msg())
                sys.exit(msg)
    # obtain image data either via PIL or CV2/numpy
    #   1. using pil
    # PIL origin (0,0) is top left corner
    pil = Image.open(imagePNGPath).convert('L') # 'L' is "black and white mode": converts to 8-bit pixels B/W
    #   2. using cv2/numpy
    #pil_1 = Image.open(imagePNGPath)
    #frame = pil_1.convert("RGB")
    #pil_gray = cv2.cvtColor(numpy.array(frame), cv2.COLOR_BGR2GRAY, dstCn=0)
    #pil = Image.fromarray(pil_gray)
    pilCropped = pil
    width, height = pil.size
    if scan_region:
        # relative (percentage) values between 0 and 1
        x_crop_min = min(scan_region[0],scan_region[2])
        x_crop_max = max(scan_region[0],scan_region[2])
        y_crop_min = min(scan_region[1],scan_region[3])
        y_crop_max = max(scan_region[1],scan_region[3])
        cropTop=int(height*y_crop_min)
        cropBottom=int(height*y_crop_max)
        cropLeft=int(height*x_crop_min)
        cropRight=int(height*x_crop_max)
        # crop box is 4-tuple: left,upper,right,lower
        pilCropBox = [cropLeft,cropTop,cropRight,cropBottom]
        pilCropped = pil.crop(pilCropBox)
    #  zbar sometimes catches a barcode at a lower resolution but
    #  misses it at a higher resolution. Scan for barcode with several
    #  variants of image specified by IMAGE_FILE_SPEC.
    barcodeString = barcode_scan_at_resolutions(pilCropped,None)
    if ( not barcodeString ):
            lg.warn(json1.json_barcode_not_found_msg([imagePNGPath],""))
    return barcodeString

def barcode_scan_at_resolutions (pil,scale_values):
    """
    Try scans at multiple image resolutions since zbar sometimes is
    befuddled by high resolution images.
    """
    if scale_values == [] :
        # done - empty array indicates all scale values have been tried
        return None;
    elif ( not scale_values ):
        # Options for leveraging zbar:
        # (1) via shell invocation and (2) via python zbar library
        #barcodeString = barcodeScan_zbarimg (pil)
        barcodeString = barcodeScan_python_zbar_sub (pil)
        if ( barcodeString ):
            return barcodeString
        else:
            scale_values = [ 0.5 ]
            return barcode_scan_at_resolutions(pil,scale_values)
    else:
        scale_value = scale_values.pop()
        resize_x = int(round(scale_value * pil.size[0]))
        resize_y = int(round(scale_value * pil.size[1]))
        pil_scaled = pil.resize( (resize_x, resize_y) )
        barcodeString = barcodeScan_python_zbar_sub (pil_scaled)
        if ( barcodeString ):
            return barcodeString
        else:
            return barcode_scan_at_resolutions(pil,scale_values)

def barcodeScan_zbarimg (pil):
    """
    If possible, return the string encoded by the barcode in the image
    specified by PIL.
    """
    barcodeString = None
    tf = tempfile.NamedTemporaryFile()
    pil.save(tf,"png")
    code_type_pairs,return_val = zbarimg(tf.name)
    lg.debug("code_type_pairs: %s",code_type_pairs)
    for code_type_pair in code_type_pairs:
        barcodeString = code_type_pair[0]
    tf.close()
    return barcodeString

def barcodeScan_python_zbar_sub (pilCropped):
    pilCroppedWidth,pilCroppedHeight = pilCropped.size
    raw = pilCropped.tobytes()
    # wrap raw image data in zbar.Image
    image = zbar.Image(pilCroppedWidth, pilCroppedHeight, 'Y800', raw)
    # create and configure a reader
    scanner = zbar.ImageScanner()
    scanner.parse_config('enable')
    # scan the image for barcodes
    scanner.scan(image)
    # extract results
    barcodeString = None
    # image.symbols should hold a zbar.SymbolSet object
    for symbol in image:
        barcodeString = symbol.data
    # clean up (destroy the image object to free up references to the data and symbols)
    # - note: if another image will be scanned, it's also possible to simply recycle the image object
    del(image)
    return barcodeString

def zbarimgWithPopen (path):
    """
    PATH can correspond to any file which the zbarimg executable can handle.
    """
    # limit to CODE128?
    # -Sdisable -Scode128.enable
    p = subprocess.Popen(['zbarimg',path],shell=False,stdout=subprocess.PIPE)
    lines = p.stdout.readlines()
    retval = p.wait()
    return parse_zbarimg_lines(lines),retval

# zbarimgWithCheckOutput
def zbarimg (path):
    """
    PATH can correspond to any file which the zbarimg executable can
    handle. Return multiple values. The first value returned is a list
    of lists; each sublist contains two members, the encoded string
    and the encoding system. The second value returned is an integer
    representing the return code (exit status) associated with
    invocation of zbarimg.
    """
    # limit to CODE128?
    # -Sdisable -Scode128.enable
    output = None
    returncode = None
    try:
        output = subprocess.check_output(
            ['zbarimg',path],
            shell=False,
            stderr=subprocess.STDOUT)
        returncode = 0
    except subprocess.CalledProcessError as e:
        returncode = e.returncode
        output = e.output
    lines = output.splitlines()
    return parse_zbarimg_lines(lines),returncode

def parse_zbarimg_line (line):
    """
    LINE is a string corresponding to a single line of zbarimg
    output. Return the encoded string and the encoding system.
    """
    # example line: 'CODE-128:1000642\n'
    line = line.rstrip()
    colon_index = line.find(':')
    return line[colon_index+1:],line[:colon_index]

def parse_zbarimg_lines (lines):
    parsed_lines = []
    for line in lines:
        parsed_line,code = parse_zbarimg_line(line)
        parsed_lines.append([parsed_line,code])
    return parsed_lines

if __name__ == "__main__":
    import sys
    # log to console when executing directly
    lg.basicConfig(stream=sys.stderr,level=logging.DEBUG)
    # only intended to be run as python ./barScan.py "/path/to/foo.png"
    lg.debug("%s",sys.argv)
    if len(sys.argv) > 1:
        barcodeScan(sys.argv[1])
    else:
        sys.exit("Must supply a single file as argument")



import argparse
import imp
import json
import math
import os
import os.path
import re
import shutil
import signal
import sys
import tempfile
import uuid

import logging
import logging.config


# define LG as the generic logger *prior* to loading any
# pdfxcb-specific modules
lg = logging
# and immediately configure the logger format
json_log_format = '%(message)s'
lg.basicConfig(format=json_log_format,stream=sys.stdout)


import pdfxcb.barScan as barScan
import pdfxcb.json1 as json1
import pdfxcb.pdf as pdf


# handle external signals requesting termination
def signal_handler(signal, frame):
    # Ensure receipt of signal is logged prior to terminating
    msg = json1.json_exit_on_external_request_msg()
    lg.error(msg)
    lg.info(json1.json_last_log_msg())
    sys.exit()

signal.signal(signal.SIGHUP, signal_handler)
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)


#
# function definitions
#
def locate_cover_sheets (png_file_tuples,containing_dir,match_re,scan_region):
    """
    Given the list of files specified by PNG_FILE_TUPLES (a set of
    tuples where the first member of each tuple specifies the name of
    the PNG file) and CONTAINING_DIR, identify those files containing
    a barcode. Return multiple values: a list of the corresponding
    barcodes and a list of the corresponding indices.
    """
    barcodes = []
    indices = []
    # I: index in IMAGE_FILES
    i = 0
    i_max = len(png_file_tuples)
    while (i<i_max):
        # log progress by default (otherwise, this can be a long period of silence...)
        lg.info(
            json1.json_progress(
                f'looking for barcode on {i} of {i_max} PNG files')
            )
        lg.info(containing_dir)
        lg.info(png_file_tuples[i][0])
        image_file_spec = os.path.join(containing_dir,png_file_tuples[i][0])
        lg.debug(image_file_spec)
        maybe_barcode = barScan.barcodeScan(
            image_file_spec,
            scan_region         # None
        )
        # don't ignore barcode if consider is true
        consider = True
        if maybe_barcode:
            if match_re:
                consider = match_re.match(maybe_barcode)
            if consider:
                barcodes.append(maybe_barcode)
                indices.append(i)
        i = i+1
        #lg.debug(barcodes)
        #lg.debug(indices)
    return barcodes,indices

def executable_sanity_checks (executables):
    """
    Check for availability of executables specified in the list of
    strings EXECUTABLES.
    """
    for executable_spec in executables:
        if not shutil.which(executable_spec):
            msg = json1.json_msg_executable_not_accessible(executable_spec)
            lg.error(msg)
            lg.info(json1.json_last_log_msg())
            sys.exit(msg)

def generate_output_file_names(cover_sheet_barcodes,
                               cover_sheet_indices,
                               output_dir):
    file_names = []
    for cover_sheet_barcode,cover_sheet_index in zip(cover_sheet_barcodes,cover_sheet_indices):
        cover_sheet_index_as_string = str.format("{0:0>03d}", cover_sheet_index)
        version = -1
        # sanity check on version (completely arbitrary at this point)
        max_version = 99
        while (not version >= max_version and
               (version < 0 or os.path.exists(path))):
            version = version + 1
            file_name = f'{cover_sheet_barcode}-{cover_sheet_index_as_string}-{version}.pdf'
            path = os.path.join(output_dir,file_name)
        file_names.append(path)
    return file_names

def generate_output_file_names_split_after(page_ranges,output_dir):
    file_names = []
    for page_range in page_ranges:
        page_range_as_string = str.format("{0:0>03d}", page_range[0]) + "-"+str.format("{0:0>03d}",page_range[1])
        version = -1
        # sanity check on version (completely arbitrary at this point)
        max_version = 99
        while (not version >= max_version and
               (version < 0 or os.path.exists(path))):
            version = version + 1
            file_name = f'{page_range_as_string}-{version}.pdf'
            path = os.path.join(output_dir,file_name)
        file_names.append(path)
    return file_names

def generate_page_ranges(cover_sheet_indices,
                         png_file_page_number_tuples,
                         number_of_pages):
    """
    Return a list of tuples. COVER_SHEET_INDICES is an array of
    integers, each an index value identifying a member of
    PNG_FILE_PAGE_NUMBER_TUPLES which corresponds to a cover sheet.
    See INVOKE_PDFIMAGES_ON docstring for more on
    PNG_FILE_PAGE_NUMBER_TUPLES. Calling code must guarantee that
    tuples in PNG_FILE_PAGE_NUMBER_TUPLES are ordered (ascending) with
    respect to page numbers.
    """
    # to capture last set of pages, tag on an imaginary cover sheet at the end
    cover_sheet_indices.append(
        len(png_file_page_number_tuples)
    )
    png_file_page_number_tuples.append((None,number_of_pages+1))
    page_ranges = []
    for cover_sheet_index, next_cover_sheet_index in zip(cover_sheet_indices[:-1],cover_sheet_indices[1:]):
        page_ranges.append(
            (png_file_page_number_tuples[cover_sheet_index][1],
             png_file_page_number_tuples[next_cover_sheet_index][1]-1))
    return page_ranges

def generate_page_ranges_split_after(split_after,number_of_pages):
    """
    Return a list of tuples.
    """
    number_of_docs = int(math.ceil(number_of_pages/split_after))
    page_ranges = [(n*split_after-(split_after-1),n*split_after) for n in range(1,number_of_docs+1)]
    # adjust the last tuple since the number of pages may not be a
    # multiple of SPLIT_AFTER (the last set of pages may be smaller
    # than split_after pages)
    last_tuple = page_ranges[len(page_ranges)-1]
    page_ranges[len(page_ranges)-1] = (last_tuple[0],number_of_pages)
    return page_ranges

def pdfxcb_sanity_checks (output_dir,pdf_file_spec,rasterize_p,region):
    # file and dir sanity checks
    if (not output_dir):
        sys.exit("The output directory must be specified.")
    directory_sanity_checks ([output_dir],True)
    file_sanity_checks ([pdf_file_spec],True)
    # region/rasterize sanity check
    if (not rasterize_p and region):
        sys.exit("If REGION is specified, then RASTERIZE_P should be true.")
    # executables sanity check
    required_executables = [
        'gs'
        #'pdftoppm'
    ]
    executable_sanity_checks(required_executables)
    required_modules = [
        'PyPDF2'
        #'cv2'
    ]
    module_sanity_checks (required_modules,True)

def file_and_dir_sanity_checks (dirs,files):
    directory_sanity_checks (dirs,True)
    file_sanity_checks (files,True)

def pdfxcb (pdf_file_spec,output_dir,match_re,rasterize_p,region,
            clean_up_png_files_p=True
            ):
    """
    Given the file specified by PDF_FILE_SPEC, look for cover sheets
    and split the PDF at each coversheet. Name output file(s) based on
    cover sheet content. Write files to directory specified by
    OUTPUT_DIR. Return True. If MATCH_RE is defined, ignore barcodes
    unless the corresponding string matches the regex MATCH_RE. Use
    RASTERIZE_P = False if the PDF does not contain vector graphics
    but is solely bitmap data (e.g., the PDF was generated from a
    scanned document). If REGION has the form [ float1, float2,
    float3, float4 ], use the region specified by REGION when scanning
    for a barcode or other indicator of a cover sheet.
    """
    global lg
    pdfxcb_sanity_checks(output_dir,pdf_file_spec,rasterize_p,region)
    # If confident that the PDF under analysis is derived from a scan
    # (i.e., contains only bitmap data), then the images embedded in
    # the PDF can be analyzed directly. If the PDF may contain vector
    # data on the cover sheet pages, then rasterization is indicated.
    # See doc/optimization.md for notes on time implications.

    # PNG_FILE_PAGE_NUMBER_TUPLES is an array where each member has
    # the form (<PNG file name>, <PDF page number>). There is no
    # guarantee that all pages in the original PDF document are
    # represented. Furthermore, there may be multiple PNG images per
    # PDF page -- i.e., the array might include ("flurpies.png",1) and
    # ("glurpies.png",1).

    # FIXME: consider having a single call here -- FOO -- that specializes on rasterize_p
    if rasterize_p:
        # extract PDF pages as image data (PNG files)
        png_file_page_number_tuples = split_pdf_to_png_files(pdf_file_spec,output_dir)
        # Once rasterized pages are generated, optionally scan for cue marks
        # CUE_INDICES = array where each member is an integer indicating index of member of png_file_page_number_tuples where the corresponding bitmap has a cue mark
        # cue_indices = scan_for_cue_marks(png_file_page_number_tuples) <-- use urh_corner_mean w/reasonable threshold (10? 20? 50?) for "black"
    else:
        # extract images directly from PDF
        png_file_page_number_tuples = invoke_pdfimages_on(pdf_file_spec,output_dir)
    # Code below expects png_file_page_number_tuples to be ordered with respect to page number.
    # Note that sorted default is ascending order.
    png_file_page_number_tuples = sorted(png_file_page_number_tuples,
                                         key=lambda tuple: tuple[1])
    #
    # locate cover sheets
    #
    lg.info("Locating cover sheets")
    if rasterize_p:
        # possibilities:
        # 1. png files represent rasterized pages
        if region:
            scan_region = region
        else:
            scan_region = ([0,0,0.7,0.5])
    else:
        # 2. png files represent images from PDF (via pdfimages)
        scan_region = None # None is not treated as the equivalent of ([0,0,1,1]). ([0,0,1,1]) triggers cropping by barcodeScan.
    cover_sheet_barcodes, cover_sheet_indices = locate_cover_sheets(png_file_page_number_tuples,output_dir,match_re,scan_region)
    print(cover_sheet_barcodes)
    if clean_up_png_files_p:
        for png_file_tuple in png_file_page_number_tuples:
            os.remove(os.path.join(output_dir,png_file_tuple[0]))
    pdf_length = pdf.pdf_number_of_pages(pdf_file_spec)
    page_ranges = generate_page_ranges(cover_sheet_indices,
                                       png_file_page_number_tuples,
                                       pdf_length)
    output_file_names = generate_output_file_names(cover_sheet_barcodes,
                                                   cover_sheet_indices,
                                                   output_dir)
    pdf.pdf_split(pdf_file_spec,output_file_names,page_ranges)
    lg.info(json1.json_msg(40,
             ['Analysis and burst completed'],
             False,
             files=output_file_names,
             data={
                 'barcodes': cover_sheet_barcodes,
                 'indices': cover_sheet_indices
             }
    ))
    return True

def pdfxcb_split_after (pdf_file_spec,output_dir,split_after_n_pp):
    """
    Given the file specified by PDF_FILE_SPEC, split the PDF after
    every SPLIT_AFTER pages. Name output file(s) based page ranges.
    Write files to directory specified by OUTPUT_DIR. Return True.
    """
    global lg
    file_and_dir_sanity_checks([output_dir],[pdf_file_spec])
    pdf_length = pdf.pdf_number_of_pages(pdf_file_spec)
    page_ranges = generate_page_ranges_split_after(split_after_n_pp,
                                                   pdf_length)
    output_file_names = generate_output_file_names_split_after(page_ranges,
                                                               output_dir)
    pdf.pdf_split(pdf_file_spec,output_file_names,page_ranges)
    lg.info(json1.json_msg(40,
             ['Analysis and burst completed'],
             False,
             files=output_file_names,
             data={
                 #'barcodes': cover_sheet_barcodes,
                 #'indices': cover_sheet_indices
             }
    ))
    return True

def directory_sanity_check (directory_spec,exitp):
    if not os.path.isdir(directory_spec):
        lg.error(json1.json_file_not_found(directory_spec))
        lg.info(json1.json_last_log_msg())
        if exitp:
            sys.exit("Directory " + directory_spec + " not found.")

def directory_sanity_checks (directories,exitp):
    for directory_spec in directories:
        directory_sanity_check(directory_spec,True)

def file_sanity_check (file,exitp):
    if not os.path.isfile(file):
        lg.error(json1.json_file_not_found(file))
        lg.info(json1.json_last_log_msg())
        if exitp:
            sys.exit("File " + file + " not found.")

def file_sanity_checks (files,exitp):
    for file in files:
        file_sanity_check(file,True)

def invoke_pdfimages_on (pdf_file_spec,output_dir):
    """
    Extract images in PDF file specified by PDF_FILE_SPEC into a
    series of files, each representing a single PNG image. Write files
    to directory specified by OUTPUT_DIR.

    Returns a list of tuples where each tuple has the structure
    (png_file,png_file_page_number) where png_file is a string representing the file name and png_file_page_number is an
    integer. The list is an ordered sequence with respect to page
    number - low to high.
    """
    png_file_page_number_tuples = None
    try:
        # sanity check
        if not os.path.isabs(pdf_file_spec):
            msg = "The input PDF must be specified as an absolute file path"
            lg.error(json1.json_msg(108,[msg],False,files=[pdf_file_spec]))
            sys.exit(msg)
        else:
            lg.info("png_file_page_number_tuples 0")
            png_file_page_number_tuples = pdf.pdfimages(pdf_file_spec,
                                                        output_dir)
    except Exception as e:
        lg.debug(str(e))
        msg = json1.json_failed_to_convert_pdf(e,pdf_file_spec)
        lg.error(msg)
        lg.info(json1.json_last_log_msg())
        sys.exit(msg)
    else:
        lg.info(json1.json_pdf_to_pngs_success(pdf_file_spec,
                                               None #png_files
        ))
        return png_file_page_number_tuples

def module_sanity_checks (module_names,exitp):
    """MODULE_NAMES is a sequence of strings"""
    for module_name in module_names:
        module_sanity_check (module_name,exitp)

def module_sanity_check (module_name,exitp):
    """MODULE_NAME is a string"""
    try:
        imp.find_module(module_name)
    except ImportError:
        msg = json1.json_msg_module_not_accessible(module_name)
        lg.error(msg)
        lg.info(json1.json_last_log_msg())
        if exitp:
            sys.exit(msg)

def split_pdf_to_png_files (pdf_file_spec,output_dir):
    """
    Split the PDF file specified by PDF_FILE_SPEC into a series of
    files, each representing a single page as a PNG image. Write files
    to the directory specified by OUTPUT_DIR.

    Return a list of tuples where the first member of each tuple is a
    string representing the file name and the second member of each
    tuple is the corresponding page number (page numbering begins at
    1).
    """
    png_files = None
    try:
        # sanity check
        if not os.path.isabs(pdf_file_spec):
            msg = "The input PDF must be specified as an absolute file path"
            lg.error(json1.json_msg(108,[msg],False,files=[pdf_file_spec]))
            sys.exit(msg)
        else:
            # array of (<file_name>,<page_number>) tuples
            png_specs = pdf.pdf_to_pngs(pdf_file_spec,output_dir)
    except Exception as e:
        msg = json1.json_failed_to_convert_pdf(e,pdf_file_spec)
        lg.error(msg)
        print("failed to convert PDF file(s): %s" % pdf_file_spec)
        print("e: %s" % e)
        lg.info(json1.json_last_log_msg())
        sys.exit(msg)
    else:
        lg.info(json1.json_pdf_to_pngs_success(pdf_file_spec,png_specs))
        return png_specs

def write_page_scores(page_scores, output_file):
    f = open(output_file, 'w')
    for page in page_scores:
        for row in page:
            for datum in row:
                f.write(str(datum))
                f.write(' ')
            f.write('\n')
        f.write('\n')
    f.close()

def write_paths(paths, output_file):
    f = open(output_file, 'w')
    for path in paths:
        f.write(path)
        f.write('\n')
    f.close()

def main():
    """Handle command-line invocation of pdfxcb.py."""
    global lg
    parser = argparse.ArgumentParser(description="This is pdfxcb")
    # Split after every Nth page. If N is 3, split into pp 1-3, 4-6, 7-9, etc.
    parser.add_argument("-e",
                        help="A whole number. If specified, split after every Nth page, ignoring -m and -r.",
                        action="store",
                        default=0,
                        dest="split_after_n_pp",
                        type=int
                        )
    parser.add_argument("-f",
                        # FIXME: "busca.log" shouldn't be hardcoded -- set up a default_log_path var to handle this
                        help="Absolute path to log file. If not specified, logs to busca.log.",
                        action="store",
                        dest="log_file",
                        type=str)
    # FIXME: is this mandatory? If so, this should be explicit.
    parser.add_argument("-d",
                        help="absolute path to output directory",
                        action="store",
                        dest="output_dir",
                        type=str)
    parser.add_argument("-m",
                        help="match barcodes to regex (ignore if no match)",
                        action="store",
                        dest="match_re_string",
                        type=str)
    parser.add_argument("-p",
                        help="identifier for a specific instance of pdfxcb",
                        action="store",
                        dest="identifier",
                        type=str)
    # These four values correspond to scan_region = ([0,0,0.7,0.5])
    parser.add_argument("-r",
                        help="specify region of the page to be evaluated",
                        action="store",
                        dest="region",
                        nargs=4,
                        type=float)
    parser.add_argument("-l",
                        help="integer between 0 (verbose) and 51 (terse) defining logging",
                        action="store",
                        dest="log_level",
                        type=int)
    # https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
    parser.add_argument('-v', '--version', action='version', version="0.0.3")
    parser.add_argument('--debug',
                        help="do not clean up files used during processing"
                        )
    parser.add_argument("input_files", help="one or more input (PDF) files",
                        # keep nargs as we may want to accept multiple PDFs as input at some point
                        nargs=1,
                        type=str)
    args = parser.parse_args()
    #
    # define logging (level, file, message format, ...)
    #
    log_level = args.log_level
    if isinstance(log_level, int) and log_level >= 0 and log_level <= 51:
        log_level = log_level
    else:
        # since this function doesn't necessarily exit quickly
        log_level = logging.INFO
    if args.log_file:
        logfile = args.log_file
    else:
        logfile = 'busca.log'
    json_log_format = '%(message)s'
    for handler in lg.getLogger().handlers:
        lg.getLogger().removeHandler(handler)
    formatter = logging.Formatter(json_log_format)
    # sanity check for existence of log file directory
    if (os.path.dirname(logfile) and
        not os.path.exists(os.path.dirname(logfile))):
        raise Exception(str.format("log file directory {0} not present",
                                   os.path.dirname(logfile)))
    file_handler = logging.FileHandler(logfile,'w')
    file_handler.setFormatter(formatter)
    lg.getLogger().addHandler(file_handler)
    lg.getLogger().setLevel(log_level)
    if args.identifier:
        identifier = args.identifier
    else:
        identifier = str(uuid.uuid1())
    rasterize_p = False
    if args.region:
        region = args.region
        rasterize_p = True  # if region is specified, rasterize_p should be true
        # lg.debug("args.region: %s", args.region)
    else:
        region = None
    # 1000[0-9][0-9][0-9]$ matches on tt user id
    match_re_string = args.match_re_string
    lg.debug(match_re_string)
    match_re = None
    if match_re_string:
        match_re = re.compile(match_re_string)
    pdf_file_spec = args.input_files[0]
    lg.debug(pdf_file_spec)
    lg.info(json1.json_first_log_msg(identifier, files = [pdf_file_spec] ))
    # generic debugging
    lg.debug(os.getcwd())         # current/working directory
    # might also want to import platform to get architecture, other details...
    if (args.split_after_n_pp > 0):
        pdfxcb_split_after(pdf_file_spec,args.output_dir,args.split_after_n_pp)
    else:
        try:
            pdfxcb(pdf_file_spec,
                   args.output_dir,
                   match_re,
                   rasterize_p,
                   region,
                   not args.debug #clean_up_png_files_p
                   )
        except Exception as e:
            lg.error("Crash and burn")
            lg.error(sys.exc_info()[0])
            raise
        lg.info(json1.json_last_log_msg())

if __name__ == "__main__":
    main()



# return array of means for upper-right-hand corner
def urh_corner_mean (pil_image):
    (width,height) = pil_image.size
    # current application is to detect black cue mark in URH corner (tt-cover-sheets) indicating a sheet w/a barcode -->> grab upper RH corner 20-pixel square
    im_crop = pil_image.crop([
        width - 20, #left
        1, #upper,
        width - 1, #right,
        20 #lower
    ])
    im_crop_stat = ImageStat.Stat(im_crop)
    return im_crop_stat.mean

def scan_for_cue_marks (png_specs):
    indices = []
    # FIXME: want a png_specs_index here...
    spec_index = 0
    while spec_index < png_specs.length:
        spec = png_specs[spec_index]
        # from PIL import Image
        im = Image.open(spec[0])
        if (urh_corner_mean(im) < 40):
            indices.append(spec_index)
        spec_index = spec_index + 1
    return indices



import io
import os
import pathlib
import re
import subprocess
import PyPDF2

import logging

#import pdfxcb.json1
import pdfxcb.json1 as json1


lg=logging


def pdf_number_of_pages(pdf_file):
    """
    Determine the number of pages in a PDF document. Return an integer.
    """
    pdf_path = pathlib.Path(pdf_file)
    with open(pdf_path,"rb") as f:
        reader = PyPDF2.PdfFileReader(f)
        # getNumPages can fail if the PDF, or an object therein, is
        # corrupt
        try:
            return reader.getNumPages()
        except Exception as e:
            lg.error(json1.json_msg(109,
                                    "Failure to open or parse a PDF file - possible indication of a corrupt PDF",
                                    None,
                                    file=pdf_file))
            raise e

def pdf_page_to_png(src_pdf, pagenum = 0, resolution = 72):
    """
    Return the specified PDF page as a wand.image.Image png.
    :param PyPDF2.PdfFileReader src_pdf: PDF from which to take pages.
    :param int pagenum: Page number to take.
    :param int resolution: Resolution for resulting png in DPI.
    """
    dst_pdf = PyPDF2.PdfFileWriter()
    dst_pdf.addPage(src_pdf.getPage(pagenum))

    pdf_bytes = io.BytesIO()
    dst_pdf.write(pdf_bytes)
    pdf_bytes.seek(0)

    img = Image(file = pdf_bytes, resolution = resolution)
    img.convert("png")
    return img

def pdf_split(input_pdf_file,output_files,page_ranges):
    """
    INPUT_PDF_FILE is a string representing the path to a PDF file.
    OUTPUT_FILES is a list of strings representing paths to output
    files corresponding to the specified page ranges. PAGE_RANGES is
    an array of tuples where each tuple specifies the first page and
    the last page of a given set of pages.
    """
    reader = PyPDF2.PdfFileReader(input_pdf_file)
    for output_file, page_range in zip(output_files,page_ranges):
        writer = PyPDF2.PdfFileWriter()
        pdf_split_internal(reader,writer,page_range)
        output_file = open(output_file,"wb")
        writer.write(output_file)
        output_file.close()

def pdf_split_internal (pdf_file_reader,pdf_file_writer,page_range):
    """
    The reader and writer are PyPDF2 objects. Add the pages, specified
    by PAGE_RANGE, from specified reader object to the specified
    writer object. PAGE_RANGE is a tuple where the elements are
    integers defining the first and last page (inclusive) in the page
    range. Page count begins at 1.
    """
    # Adjust page numbers as PyPDF2 counts pages beginning at zero.
    pages = list(range(page_range[0]-1,page_range[1]))
    for page_index in pages:
        pdf_file_writer.addPage(pdf_file_reader.getPage(page_index))

def pdf_to_pngs(pdf_file,output_dir):
    """
    Generate PNG files, one corresponding to each page of the PDF file
    PDF_FILE. Write files to directory specified by OUTPUT_DIR. Return
    a list of the PNG file names.
    """
    input_file_sans_suffix, input_file_suffix = os.path.splitext(pdf_file)
    maybe_dir, input_file_name_only = os.path.split(input_file_sans_suffix)
    number_of_pages = None
    outfile_root = input_file_name_only
    number_of_pages = pdf_number_of_pages(pdf_file)
    lg.info(json1.json_pdf_info(number_of_pages))
    return pdf_to_pngs__pdftoppm(pdf_file,
                                 number_of_pages,
                                 outfile_root,
                                 output_dir)

def pdf_to_pngs__gs (pdf_file, number_of_pages, outfile_root, output_dir):
    """
    Helper relying on Ghostscript. Return a list of file names.
    """
    # see common-lisp/tt-cover-sheets/conversion-of-pdf-to-png-and-zbar.txt
    pdf_to_png_res = "72"      # 72 dpi
    output_dir_and_filename = os.path.join(output_dir,outfile_root)
    # %03d is printf directive directing gs to specify page number as a zero-padded 3-digit sequence
    output_path_spec = output_dir_and_filename + '-%03d.png'
    # issue: gs doesn't give feedback regarding extent of progress
    gs_command = [
        "gs",
        "-q",
        "-dBATCH",
        "-dNOPAUSE",
        "-sDEVICE=pnggray",
        '-r'+pdf_to_png_res,
        #"-dAutoRotatePages=/PageByPage",
        '-dUseCropBox',
        # use %d as a printf format specification for page number
        # as zero-filled with minimum of three spaces
        "-sOutputFile=%s" % output_path_spec,
        pdf_file
    ]
    return_code = subprocess.call(gs_command, shell=False)
    # log success/failure
    pdf_to_pngs__gs_log(return_code,number_of_pages)
    # return file names
    return pdf_to_pngs__gs_file_names (number_of_pages,outfile_root)

def pdf_to_pngs__gs_log (return_code,number_of_pages):
    if (return_code == 0):
        for page_number in range(number_of_pages):
            lg.info(json1.json_completed_pdf_to_ppm(page_number,number_of_pages))
    else:
        lg.error(json1.json_failed_to_convert_pdf(None,pdf_file))

def pdf_to_pngs__gs_file_names (number_of_pages,outfile_root):
    png_files=[]
    index_format_string = "{1:0>03d}"
    string_format_string = "{0}-" + index_format_string + ".png"
    for pagenumber in range(number_of_pages):
        png_infile = str.format(
            string_format_string,
            outfile_root,pagenumber+1);
        png_files.append(png_infile)
    return png_files

def pdf_to_pngs__pdftoppm (pdf_file, number_of_pages, outfile_root, output_dir):
    """
    Helper relying on pdftoppm. OUTFILE_ROOT is the filename only (no
    directory information). Return a list where each member has the
    form (<file name>,<page number>) with page numbering beginning at
    one.
    """
    output_dir_and_filename = os.path.join(output_dir,outfile_root)
    # Q: what is the point of calling pdftoppm repeatedly on each page, one at a time? doesn't this generate the exact same set of files as calling it just once w/o the -f and -l options?
    for page_number in range(number_of_pages):
        returncode = subprocess.call(
            ["pdftoppm", "-f", str(page_number+1),
             "-l", str(page_number+1),
             "-gray",
             "-png",
             pdf_file,
             output_dir_and_filename],
            shell=False)
        if (returncode == 0):
            lg.info(json1.json_completed_pdf_to_ppm(page_number,number_of_pages))
        else:
            lg.error(json1.json_failed_to_convert_pdf(None,pdf_file))
    # Return an array where each member has the form
    # (<file name>,<page number>)
    return_value = []
    # Due to the inability to configure the output file name format
    # for pdftoppm, plan ahead for the file names, anticipating
    # pdftoppm's default non-configurable behavior.
    index_format_string = ""
    if ( number_of_pages < 10 ):
        index_format_string = "{1:d}"
    elif ( number_of_pages < 100 ):
        index_format_string = "{1:0>02d}"
    elif ( number_of_pages < 1000 ):
        index_format_string = "{1:0>03d}"
    else:
        raise Exception('no support (at this point) for page count exceeding 1000 pages')
    string_format_string = "{0}-" + index_format_string + ".png"
    for pagenumber in range(number_of_pages):
        png_file = str.format(
            string_format_string,
            outfile_root, # output_dir_and_filename
            pagenumber+1);
        return_value.append((png_file,
                             pagenumber+1
                             ))
    return return_value

def pdfimages(pdf_file,output_dir):
    """
    Generate PNG files, one corresponding to each image in the PDF
    file PDF_FILE. Write files to directory specified by OUTPUT_DIR.

    Return tuples where each member has the form (png-file-name,
    page-number) where png-file-name is a string representing the name
    of the file and page-number represents the page number in the
    corresponding PDF file. Page numbering begins at 1.
    """
    input_file_sans_suffix, input_file_suffix = os.path.splitext(pdf_file)
    maybe_dir, input_file_name_only = os.path.split(input_file_sans_suffix)
    outfile_root = input_file_name_only
    output_dir_and_filename = os.path.join(output_dir,outfile_root)
    returncode = subprocess.call(
        ["pdfimages", "-p", "-png", pdf_file, output_dir_and_filename],
        shell=False)
    if (returncode == 0):
        # FIXME: this is a problem if other programs rely on this -- should be in docstring if it's guaranteed to log this
        #lg.info(json1.json_completed_pdf_to_ppm(page_number,number_of_pages))
        lg.info(json1.json_completed_pdf_to_ppm(-1,-1))
    else:
        lg.error(json1.json_failed_to_convert_pdf(None,pdf_file))
    # return values
    png_file_page_number_tuples=[]
    # file names have the form <image root>-<page number>-<image number>.png where the numbers are 3-digit zero-padded values
    # - it would be great if pdfimages, w/o a single invocation, could (1) extract images *and* (2) provide list of images
    dir_files = os.listdir(output_dir)
    outfile_root_re = re.compile("^"+outfile_root+"-(\d{1,3}\d{1,3}\d{1,3})-\d{1,3}\d{1,3}\d{1,3}\.png$")
    for dir_file in dir_files:
        png_file_match = outfile_root_re.match(dir_file)
        if png_file_match:
            png_file_page_number_tuples.append(
                ( png_file_match.group(),
                  int(png_file_match.group(1))
                )
            )
    return png_file_page_number_tuples



import json
import datetime
import time


def json_msg(code,message,outfile,
             data=None,file=None,files=None,pdffile=None,pngfiles=None):
    """
    Return a string or, if OUTFILE is a string, send to the file
    corresponding to OUTFILE. The output should be a JSON object with a
    code slot, a message slot, a time slot, a microsec slot, and
    additional slots as specified by the function's additional
    arguments/parameters.

    If message is an array, the first member of MESSAGE should be
    succinct and not disclose excessive detail. The intent is that
    this string could serve as an update for a non-technical end-user.
    """
    # if appendp is True (the default), create file if nonexistent; append if existent
    appendp=True
    if appendp:
        openMode='a+'
    else:
        openMode='w+'
    obj = json_msg_obj(code,
                       message,
                       data=data,
                       file=file,
                       files=files,
                       pdffile=pdffile,
                       pngfiles=pngfiles)
    if outfile:
        with open(outfile, openMode) as outfp:
            json.dump(obj, outfp)
    else:
        return json.dumps(obj)

def json_msg_obj(code,message,
                 data=None,file=None,files=None,pdffile=None,pngfiles=None):
    """Return an object."""
    obj = {}
    obj['code'] = code
    obj['message'] = message
    obj['time'] = int(time.time())
    obj['microsec'] = datetime.datetime.now().microsecond
    if data:
        obj['data'] = data
    if file:
        obj['file'] = file
    if files:
        obj['files'] = files
    if pdffile:
        obj['pdffile'] = pdffile
    if pngfiles:
        obj['pngfiles'] = pngfiles
    return obj

#
# messages
#
def json_barcode_not_found_msg(files,msg):
    """
    FILES is an array where each member is a string specifying the
    location of a diagnostic image file. MSG is additional data
    encapsulated as a string.
    """
    message = ['Barcode was not found',
               'see diagnostic image file(s) at {}'.format(files)
               ]
    if msg:
        message.append(msg);
    return json_msg(135, message, False,None)

def json_blank_page_on_deskew(file):
    """Return a string"""
    return json_msg(121,
                    "encountered blank page on attempt to deskew",
                    False,file=file)

def json_completed_pdf_to_ppm(page_number,number_of_pages):
    """Return a string"""
    return json_msg(11,
                    'Completed PDF to PPM conversion: page {} / {}'.format(page_number,number_of_pages),
                    False,None)

def json_converting_pdf(file):
    """Return a string"""
    return json_msg(0,
                    "Converting the PDF to PNG images... this may take some time...",
                    False,file=file)

def json_directory_not_found(dir):
    return json_msg(136,
                    'Directory not found; directory: {}'.format(dir),
                    False,None)

def json_first_log_msg(identifier,files=None):
    """Return a string. Use for the first log message."""
    obj = json_msg_obj(3,"Logging initiated")
    obj['id'] = identifier
    obj['files'] = files
    return json.dumps(obj)

def json_last_log_msg():
    """Return a string. Use for the last log message."""
    return json_msg(2,
                    "Scan and analysis complete",
                    False,None)

def json_msg_executable_not_accessible(executable_name):
    """EXECUTABLE_NAME is a string."""
    return json_msg(134,
             'The executable ' + executable_name + ' is not accessible. Is it installed?',
             False,None)

def json_exit_on_external_request_msg():
    return json_msg(4,
                    "Received external request to terminate process",
                    False,None)

def json_failed_to_convert_pdf(exception,PDFFileSpec):
    return json_msg(110,
                    ['Failed to convert PDF to PNG(s)', str(exception)],
                    False,
                    files=PDFFileSpec)

# ideally, pngFile specifies the absolute path to an image with as
# much diagnostic or explanatory information as possible to assist the
# end-user in understanding why the deskew failed
def json_failed_to_deskew(pngFile,pageNumber,comments):
    message = 'Failed to deskew at page {}'.format(pageNumber)
    message = message + " " + comments
    return json_msg(120,
                    message,
                    False,
                    files=[pngFile])

def json_failed_to_parse_file(exception,someFile):
    return json_msg(131,
                    ['Failed to parse file', str(exception)],
                    False,file=someFile)

def json_file_not_found(file):
    return json_msg(132,
                    'File not found; file: {}'.format(file),
                    False,None)

# FILES is an array where each member is a string specifying the location of a diagnostic image file. This is only generated if debug_p flag is set in code of interest.

# RECT should be an array with two points [(x1,y1),(x2,y2)] defining a rectangle centered on and surrounding the area of interest. DIM should be an array with two values, the width (x) and the length (y) of the image. Units for components of DIM should be the same units as those used for components of RECT. PAGE_N is the page number relative to the document specified by PDF_FILE.

# this log message should allow any client, given the PDF file under consideration, to generate a diagnostic image
def json_msg_bubble_not_found(files,msg,rect,dim,page_n):
    """MSG is additional data encapsulated as a string"""
    data = { "files": files,
             "area":  rect,
             "dimensions": dim,
             "page_n": page_n
    }
    message = ['anticipated a bubble at position but bubble was not found', 'see diagnostic image files']
    if msg:
        message.append(msg)
    return json_msg(133, message, False, data=data)

def json_msg_bubbles_not_found(file,msg):
    """
    FILE is a single string specifying the location of a diagnostic
    image file. MSG is additional data encapsulated as a string. This
    is intended to be used to provide summary information to the end
    user. Individual bubble issues should be logged with
    json_msg_bubble_not_found.
    """
    message = ['Unable to find one or more bubbles', 'see diagnostic image file']
    if msg:
        message.append(msg)
    return json_msg(134, message, False, file=file)

def json_msg_module_not_accessible(module_name):
    """MODULE_NAME is a string."""
    return json_msg(140,
                    'The python module ' + module_name + ' is not accessible. Is it installed?',
                    False,None)

# File size (kB), resolution, file name, etc. might also be of interest at some point.
def json_pdf_info(number_of_pages):
    """Provide description of the PDF under consideration."""
    pdf_data = { 'number_of_pages': number_of_pages }
    return json_msg(70,
                    "PDF information",
                    False, data=pdf_data)

def json_pdf_to_pngs_success(pdffile,png_specs):
    """
    Return a string. PNG_SPECS is an array of (<file_name>,<page_number>)
    tuples.
    """
    return json_msg(10,
                    "Successfully converted PDF to PNG(s)",
                    False,
                    pdffile=pdffile)

def json_progress(progress_message):
    """
    Use to provide an informational message indicating extent of
    progress.
    """
    return json_msg(50, progress_message, False)

def json_scanset(scanSet):
    return json_msg(30,
                    'scanset',
                    False, data=scanSet);

def json_scansets(scanSets):
    return json.dumps(scanSets)

def json_successful_deskew(file):
    """Return a string"""
    return json_msg(20,
                    "Successful deskew",
                    False, file=file)
